import tensorflow as tf


class CircularLearningRate(tf.keras.callbacks.Callback):
    def __init__(self, min_lr_1, max_lr, min_lr_2, data, batch_size):
        super().__init__()
        self.min_lr_1 = min_lr_1
        self.max_lr = max_lr
        self.min_lr_2 = min_lr_2
        self.mid = len(data) / batch_size / 2

    def on_train_batch_end(self, batch, logs=None):
        if batch < self.mid:
            lb = self.min_lr_1
            way = self.mid - batch
        else:
            lb = self.min_lr_2
            way = batch - self.mid
        lr = self.max_lr - way / self.mid * (self.max_lr - lb)
        tf.keras.backend.set_value(self.model.optimizer.lr, lr)


class CircularLearningRateSched(tf.keras.optimizers.schedules.LearningRateSchedule):
    def __init__(self, min_lr_1, max_lr, min_lr_2, mid=1000):
        super().__init__()
        self.min_lr_1 = min_lr_1
        self.max_lr = max_lr
        self.min_lr_2 = min_lr_2
        self.mid = mid
    
    @tf.function
    def __call__(self, step):
        if step < self.mid:
            lb = self.min_lr_1
            way = self.mid - step
        else:
            lb = self.min_lr_2
            way = step - self.mid
        return self.max_lr - way / self.mid * (self.max_lr - lb)



import tensorflow as tf

from nakdimon.train import TrainingParams
from nakdimon import schedulers


class FullTraining(TrainingParams):
    pass


class SingleLayer(TrainingParams):
    def build_model(self):
        from train import LETTERS_SIZE, NIQQUD_SIZE, DAGESH_SIZE, SIN_SIZE
        layers = tf.keras.layers

        inp = tf.keras.Input(shape=(None,), batch_size=None)
        embed = layers.Embedding(LETTERS_SIZE, self.units, mask_zero=True)(inp)

        layer = layers.Bidirectional(layers.LSTM(self.units, return_sequences=True, dropout=0.1), merge_mode='sum')(embed)
        layer = layers.Dense(self.units)(layer)

        outputs = [
            layers.Dense(NIQQUD_SIZE, name='N')(layer),
            layers.Dense(DAGESH_SIZE, name='D')(layer),
            layers.Dense(SIN_SIZE, name='S')(layer),
        ]
        return tf.keras.Model(inputs=inp, outputs=outputs)


def SingleLayerLarge():
    return SingleLayer(557)


class SplitSin(TrainingParams):
    def build_model(self):
        from nakdimon.train import LETTERS_SIZE, NIQQUD_SIZE, DAGESH_SIZE, SIN_SIZE
        layers = tf.keras.layers

        inp = tf.keras.Input(shape=(None,), batch_size=None)
        embed = layers.Embedding(LETTERS_SIZE, self.units, mask_zero=True)(inp)

        layer = layers.Bidirectional(layers.LSTM(self.units, return_sequences=True, dropout=0.1), merge_mode='sum')(embed)
        layer = layers.Bidirectional(layers.LSTM(self.units, return_sequences=True, dropout=0.1), merge_mode='sum')(layer)
        layer = layers.Dense(self.units)(layer)

        sin_layer = layers.Bidirectional(layers.LSTM(self.units, return_sequences=True, dropout=0.1), merge_mode='sum')(embed)

        outputs = [
            layers.Dense(NIQQUD_SIZE, name='N')(layer),
            layers.Dense(DAGESH_SIZE, name='D')(layer),
            layers.Dense(SIN_SIZE, name='S')(sin_layer),
        ]
        return tf.keras.Model(inputs=inp, outputs=outputs)


class UnmaskedLoss(TrainingParams):
    def loss(self, y_true, y_pred):
        return tf.keras.losses.sparse_categorical_crossentropy(y_true, y_pred, from_logits=True)


class ConstantLR(TrainingParams):
    def __init__(self, lr_string):
        super().__init__()
        self.lr = float(lr_string)
        self.lr_string = lr_string

    @property
    def name(self):
        return f'ConstantLR({self.lr_string})'

    def epoch_params(self, data):
        scheduler = tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: self.lr)
        yield ('premodern', 1, scheduler)
        yield ('modern', 5, tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: self.lr))


class ModernOnly(TrainingParams):
    corpus = {
        'modern': [
            'hebrew_diacritized/modern',
            'hebrew_diacritized/dictaTestCorpus'
        ]
    }

    def epoch_params(self, data):
        lrs = [30e-4, 30e-4, 30e-4, 8e-4, 1e-4]
        yield ('modern', len(lrs), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs[epoch]))


class Quick(TrainingParams):
    def epoch_params(self, data):
        yield ('modern', 1, tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: 3e-3))


class Chunk(TrainingParams):
    def __init__(self, maxlen):
        super().__init__()
        self.maxlen = int(maxlen)

    @property
    def name(self):
        return f'Chunk({self.maxlen})'


class Batch(TrainingParams):
    def __init__(self, batch_size):
        super().__init__()
        self.batch_size = int(batch_size)

    @property
    def name(self):
        return f'Batch({self.batch_size})'


class FullNoMix(TrainingParams):
    corpus = {
        'automatic': tuple([
            'hebrew_diacritized/shortstoryproject_predotted',
            'hebrew_diacritized/shortstoryproject_Dicta',
        ]),
        'modern': tuple([
            'hebrew_diacritized/modern',
            'hebrew_diacritized/dictaTestCorpus'
        ])
    }

    def epoch_params(self, data):
        lrs1 = [30e-4, 10e-4]
        yield ('automatic', len(lrs1), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs1[epoch-1]))
        lrs2 = [10e-4, 10e-4, 3e-4]
        yield ('modern', len(lrs2), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs2[epoch-len(lrs1)-1]))


class FullOrdered(TrainingParams):
    corpus = {
        'rabanit': tuple([
            'hebrew_diacritized/rabanit',
        ]),
        'pre_modern': tuple([
            'hebrew_diacritized/pre_modern',
            'hebrew_diacritized/shortstoryproject_predotted',
        ]),
        'automatic': tuple([
            'hebrew_diacritized/shortstoryproject_Dicta',
        ]),
        'modern': tuple([
            'hebrew_diacritized/modern',
            'hebrew_diacritized/dictaTestCorpus'
        ])
    }

    def epoch_params(self, data):
        yield ('rabanit', 1, tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: 30e-4))
        yield ('pre_modern', 1, tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: 30e-4))
        yield ('automatic', 1, tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: 30e-4))
        yield ('modern', 3, tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: 30e-4))


class NoPredotted(TrainingParams):
    corpus = {
        'automatic': tuple([
            'hebrew_diacritized/shortstoryproject_predotted',
            'hebrew_diacritized/shortstoryproject_Dicta',
        ]),
        'modern': tuple([
            'hebrew_diacritized/modern',
            'hebrew_diacritized/dictaTestCorpus'
        ])
    }

    def epoch_params(self, data):
        lrs1 = [30e-4, 10e-4]
        yield ('automatic', len(lrs1), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs1[epoch-1]))
        lrs2 = [10e-4, 10e-4, 3e-4]
        yield ('modern', len(lrs2), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs2[epoch-len(lrs1)-1]))


class FullUpdated(TrainingParams):
    corpus = {
        'premodern': tuple([
            'hebrew_diacritized/poetry',
            'hebrew_diacritized/rabanit',
            'hebrew_diacritized/pre_modern',
            'hebrew_diacritized/shortstoryproject_predotted'
        ]),
        'automatic': tuple([
            'hebrew_diacritized/shortstoryproject_Dicta',
        ]),
        'modern': tuple([
            'hebrew_diacritized/modern',
            'hebrew_diacritized/dictaTestCorpus',
            'hebrew_diacritized/new'
        ])
    }

    def epoch_params(self, data):
        yield ('premodern', 1, schedulers.CircularLearningRate(3e-3, 8e-3, 1e-4, data['premodern'][0], self.batch_size))
        lrs1 = [30e-4, 10e-4]
        yield ('automatic', len(lrs1), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs1[epoch-1]))
        lrs2 = [10e-4, 10e-4, 3e-4]
        yield ('modern', len(lrs2), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs2[epoch-len(lrs1)-1]))


class TasteModernFirst(FullUpdated):
    def epoch_params(self, data):
        yield ('modern', 1,  tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: 3e-3))
        yield ('premodern', 1, schedulers.CircularLearningRate(3e-3, 8e-3, 1e-4, data['premodern'][0], self.batch_size))
        lrs1 = [30e-4, 10e-4]
        yield ('automatic', len(lrs1), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs1[epoch-1-1]))
        lrs2 = [10e-4, 10e-4, 3e-4]
        yield ('modern', len(lrs2), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs2[epoch-len(lrs1)-1-1]))



from __future__ import annotations
import itertools
from collections import defaultdict, Counter
from dataclasses import dataclass
from typing import NamedTuple, Iterator, Iterable, List, Tuple
from functools import lru_cache
import re

from nakdimon import utils


# "rafe" denotes a letter to which it would have been valid to add a diacritic of some category
# but instead it is decided not to. This makes the metrics less biased.
RAFE = '\u05BF'


class Niqqud:
    SHVA = '\u05B0'
    REDUCED_SEGOL = '\u05B1'
    REDUCED_PATAKH = '\u05B2'
    REDUCED_KAMATZ = '\u05B3'
    HIRIK = '\u05B4'
    TZEIRE = '\u05B5'
    SEGOL = '\u05B6'
    PATAKH = '\u05B7'
    KAMATZ = '\u05B8'
    HOLAM = '\u05B9'
    KUBUTZ = '\u05BB'
    SHURUK = '\u05BC'
    METEG = '\u05BD'


HEBREW_LETTERS = [chr(c) for c in range(0x05d0, 0x05ea + 1)]

NIQQUD = [RAFE] + [chr(c) for c in range(0x05b0, 0x05bc + 1)] + ['\u05b7']

HOLAM = Niqqud.HOLAM

SHIN_YEMANIT = '\u05c1'
SHIN_SMALIT = '\u05c2'
NIQQUD_SIN = [RAFE, SHIN_YEMANIT, SHIN_SMALIT]  # RAFE is for acronyms

DAGESH_LETTER = '\u05bc'
DAGESH = [RAFE, DAGESH_LETTER]  # note that DAGESH and SHURUK are one and the same

ANY_NIQQUD = [RAFE] + NIQQUD[1:] + NIQQUD_SIN[1:] + DAGESH[1:]

VALID_LETTERS = [' ', '!', '"', "'", '(', ')', ',', '-', '.', ':', ';', '?'] + HEBREW_LETTERS
SPECIAL_TOKENS = ['H', 'O', '5']

ENDINGS_TO_REGULAR = dict(zip('ךםןףץ', 'כמנפצ'))


def normalize(c):
    if c in VALID_LETTERS: return c
    if c in ENDINGS_TO_REGULAR: return ENDINGS_TO_REGULAR[c]
    if c in ['\n', '\t']: return ' '
    if c in ['־', '‒', '–', '—', '―', '−']: return '-'
    if c == '[': return '('
    if c == ']': return ')'
    if c in ['´', '‘', '’']: return "'"
    if c in ['“', '”', '״']: return '"'
    if c.isdigit(): return '5'
    if c == '…': return ','
    if c in ['ײ', 'װ', 'ױ']: return 'H'
    return 'O'


class HebrewItem(NamedTuple):
    letter: str
    normalized: str
    dagesh: str
    sin: str
    niqqud: str

    def __str__(self):
        return self.letter + self.dagesh + self.sin + self.niqqud

    def __repr__(self):
        return repr((self.letter, bool(self.dagesh), bool(self.sin), ord(self.niqqud or chr(0))))

    def vocalize(self):
        return self._replace(niqqud=vocalize_niqqud(self.niqqud),
                             sin=self.sin.replace(RAFE, ''),
                             dagesh=vocalize_dagesh(self.letter, self.dagesh))

    def to_text(self) -> str:
        return str(self).replace(RAFE, '')


def items_to_text(items: List[HebrewItem]) -> str:
    return ''.join(str(item) for item in items).replace(RAFE, '')


def vocalize_dagesh(letter, dagesh):
    if letter not in 'בכפ':
        return ''
    return dagesh.replace(RAFE, '')


def vocalize_niqqud(c):
    # FIX: HOLAM / KUBBUTZ cannot be handled here correctly
    if c in [Niqqud.KAMATZ, Niqqud.PATAKH, Niqqud.REDUCED_PATAKH]:
        return Niqqud.PATAKH

    if c in [Niqqud.HOLAM, Niqqud.REDUCED_KAMATZ]:
        return Niqqud.HOLAM  # TODO: Kamatz-katan

    if c in [Niqqud.SHURUK, Niqqud.KUBUTZ]:
        return Niqqud.KUBUTZ

    if c in [Niqqud.TZEIRE, Niqqud.SEGOL, Niqqud.REDUCED_SEGOL]:
        return Niqqud.SEGOL

    if c == Niqqud.SHVA:
        return ''

    return c.replace(RAFE, '')


def is_hebrew_letter(letter: str) -> bool:
    return '\u05d0' <= letter <= '\u05ea'


def is_hebrew_niqqud(letter: str) -> bool:
    return letter in ANY_NIQQUD


def can_dagesh(letter):
    return letter in ('בגדהוזטיכלמנספצקשת' + 'ךף')


def can_sin(letter):
    return letter == 'ש'


def can_niqqud(letter):
    return letter in ('אבגדהוזחטיכלמנסעפצקרשת' + 'ךן')


def can_any(letter):
    return can_niqqud(letter) or can_dagesh(letter) or can_sin(letter)


def iterate_dotted_text(text: str) -> Iterator[HebrewItem]:
    n = len(text)
    text += '  '
    i = 0
    while i < n:
        letter = text[i]

        dagesh = RAFE if can_dagesh(letter) else ''
        sin = RAFE if can_sin(letter) else ''
        niqqud = RAFE if can_niqqud(letter) else ''
        normalized = normalize(letter)
        i += 1
        nbrd = text[i - 15:i + 15].split()[1:-1]

        assert letter not in ANY_NIQQUD, f'{i}, {nbrd}, {letter}, {[name_of(c) for word in nbrd for c in word]}'

        if is_hebrew_letter(normalized):
            if text[i] == DAGESH_LETTER:
                # assert dagesh == RAFE, (text[i-5:i+5])
                dagesh = text[i]
                i += 1
            if text[i] in NIQQUD_SIN:
                # assert sin == RAFE, (text[i-5:i+5])
                sin = text[i]
                i += 1
            if text[i] in NIQQUD:
                # assert niqqud == RAFE, (text[i-5:i+5])
                niqqud = text[i]
                i += 1
            if letter == 'ו' and dagesh == DAGESH_LETTER and niqqud == RAFE:
                dagesh = RAFE
                niqqud = DAGESH_LETTER

        yield HebrewItem(letter, normalized, dagesh, sin, niqqud)


def iterate_file(path) -> Iterator[HebrewItem]:
    with open(path, encoding='utf-8') as f:
        text = ''.join(s + ' ' for s in f.read().split())
        try:
            yield from iterate_dotted_text(text)
        except AssertionError as ex:
            ex.args += (path,)
            raise


def is_space(c) -> bool:
    if isinstance(c, HebrewItem):
        return c.letter == ' '
    elif isinstance(c, str):
        return c == ' '
    assert False


def split_by_length(characters: Iterable, maxlen: int):
    assert maxlen > 1
    out = []
    space = maxlen
    for c in characters:
        if is_space(c):
            space = len(out)
        out.append(c)
        if len(out) == maxlen - 1:
            yield out[:space+1]
            out = out[space+1:]
    if out:
        yield out


@dataclass(frozen=True)
class Token:
    items: tuple[HebrewItem, ...]

    def __str__(self):
        return ''.join(str(c) for c in self.items)

    def __lt__(self, other: 'Token'):
        return (self.to_undotted(), str(self)) < (other.to_undotted(), str(other))

    def split_on_hebrew(self) -> tuple[str, 'Token', str]:
        start = 0
        end = len(self.items) - 1
        while True:
            if start >= len(self.items):
                return ('', Token(()), '')
            if self.items[start].letter in HEBREW_LETTERS + ANY_NIQQUD:
                break
            start += 1
        while self.items[end].letter not in HEBREW_LETTERS + ANY_NIQQUD:
            end -= 1
        return (''.join(c.letter for c in self.items[:start]),
                Token(self.items[start:end+1]),
                ''.join(c.letter for c in self.items[end+1:]))

    def __bool__(self):
        return bool(self.items)

    def __eq__(self, other):
        return self.items == other.items

    @lru_cache()
    def to_undotted(self):
        return ''.join(str(c.letter) for c in self.items)

    def is_undotted(self):
        return len(self.items) > 1 and all(c.niqqud in [RAFE, ''] for c in self.items)

    def is_definite(self):
        return len(self.items) > 2 and self.items[0].niqqud == 'הַ'[-1] and self.items[0].letter in 'כבלה'

    def vocalize(self) -> 'Token':
        return Token(tuple([c.vocalize() for c in self.items]))

    def is_hebrew(self) -> bool:
        return len([c for c in self.items if c.letter in HEBREW_LETTERS]) > 1

    def to_text(self) -> str:
        return str(self).replace(RAFE, '')


def tokenize(char_iterator: Iterator[HebrewItem], strip_nonhebrew: bool) -> list[Token]:
    result = []
    current = []
    for c in char_iterator:
        if c.letter.isspace() or c.letter == '-':
            if current:
                token = Token(tuple(current))
                if strip_nonhebrew:
                    _, token, _ = token.split_on_hebrew()
                result.append(token)
            current = []
        else:
            current.append(c)
    if current:
        token = Token(tuple(current))
        if strip_nonhebrew:
            _, token, _ = token.split_on_hebrew()
        result.append(token)
    return result


def collect_wordmap(tokens: Iterable[Token]):
    word_dict = defaultdict(Counter)
    for token in tokens:
        word_dict[token.to_undotted()][str(token)] += 1
    return word_dict


def collect_tokens(paths: Iterable[str]) -> list[Token]:
    return tokenize(itertools.chain.from_iterable(iterate_file(path) for path in utils.iterate_files(paths)),
                    strip_nonhebrew=True)


def count_hebrew_tokens(paths: Iterable[str]):
    return sum(1 for token in collect_tokens(paths) if token.is_hebrew())


def stuff(tokens):
    stripped_tokens = [token.strip_nonhebrew() for token in tokens if token.strip_nonhebrew()]
    word_dict = collect_wordmap(stripped_tokens)
    # for k, v in sorted(word_dict.items(), key=lambda kv: (len(kv[1]), sum(kv[1].values()))):
    #     print(k, ':', str(v).replace('Counter', ''))
    # print(len(word_dict))

    for t in tokens:
        if t.is_definite() and t.items[1].letter not in 'אהחער' and not t.items[1].dagesh:
            print(t)
    #
    # for t in stripped_tokens:
    #     if t.is_undotted() and '"' not in t.to_undotted():
    #         print(t)
    # for k, v in word_dict.items():
    #     if "וו" in k:
    #         print(v)


def remove_niqqud(text: str) -> str:
    return re.sub('[\u05B0-\u05BC\u05C1\u05C2ׇ\u05c7\u05BF]', '', text)


def name_of(c):
    if 'א' <= c <= 'ת':
        return c
    if c == DAGESH_LETTER: return 'דגש\שורוק'
    if c == Niqqud.KAMATZ: return 'קמץ'
    if c == Niqqud.PATAKH: return 'פתח'
    if c == Niqqud.TZEIRE: return 'צירה'
    if c == Niqqud.SEGOL: return 'סגול'
    if c == Niqqud.SHVA: return 'שוא'
    if c == Niqqud.HOLAM: return 'חולם'
    if c == Niqqud.KUBUTZ: return 'קובוץ'
    if c == Niqqud.HIRIK: return 'חיריק'
    if c == Niqqud.REDUCED_KAMATZ: return 'חטף-קמץ'
    if c == Niqqud.REDUCED_PATAKH: return 'חטף-פתח'
    if c == Niqqud.REDUCED_SEGOL: return 'חטף-סגול'
    if c == SHIN_SMALIT: return 'שין-שמאלית'
    if c == SHIN_YEMANIT: return 'שין-ימנית'
    if c.isprintable():
        return c
    return "לא ידוע ({})".format(hex(ord(c)))


def find_longest_undotted(text):
    text = re.sub(r'[^\u0590-\u05f4#]', '', text)
    return max(re.findall(r'[\u05d0-\u05ea]+', text), key=len)


def print_longest_undotted_files(path):
    undotted_list = [(find_longest_undotted(utils.read_file(filename)), filename) for filename in utils.iterate_files(path)]
    undotted_list.sort(key=lambda x: len(x[1]))
    for text, filename in undotted_list:
        if len(text) > 25:
            print(filename)


def lsplit_nonhebrew(word: str) -> tuple[str, str]:
    assert word
    i = 0
    for i in range(len(word)):
        if is_hebrew_letter(word[i]) or is_hebrew_niqqud(word[i]):
            break
    return word[:i], word[i:]


def rsplit_nonhebrew(word: str) -> tuple[str, str]:
    assert word
    right, word = lsplit_nonhebrew(word[::-1])
    return word[::-1], right[::-1]


def split_nonhebrew(word: str) -> tuple[str, str, str]:
    assert word
    left, word = lsplit_nonhebrew(word)
    word, right = rsplit_nonhebrew(word)
    return left, word, right


def table_2():
    from pathlib import Path
    USER_GEN = r"{*}         & User-gen.  & Blogs, forums "
    NEWS_MAG = r"{*}         & News / Mag & Online outlets"
    BOOKS =    r"{$\dagger$} & Literary   & Books, forums "
    WIKI =     r"{*}         & Wiki       & he.wikipedia  "
    OFFICIAL = r"{*}         & Official   & gov.il        "
    YANSHUF =  r"            & News       & Yanshuf       "
    DICTA =    r"            & Wiki       & Dicta test set"
    genre = {
        "dictaTestCorpus": DICTA,
        "blogs": USER_GEN,
        "critics": USER_GEN,
        "forums": USER_GEN,
        "news": NEWS_MAG,
        "ynet": NEWS_MAG,
        "bsheva": NEWS_MAG,
        "yisrael-hayom": NEWS_MAG,
        "hadrei-haredim": NEWS_MAG,
        "random_internet": BOOKS,
        "books": BOOKS,
        "dont_panic": BOOKS,
        "wiki": WIKI,
        "yanshuf": YANSHUF,
        "govil": OFFICIAL,
        "kol-briut": OFFICIAL,
    }
    docs = Counter()
    total_docs = 0

    tokens = Counter()
    total_tokens = 0

    for path in itertools.chain(Path('hebrew_diacritized/modern').iterdir(),
                                Path('hebrew_diacritized/dictaTestCorpus').iterdir(),
                                Path('hebrew_diacritized/validation/modern').iterdir(),
                                Path('hebrew_diacritized/validation/dictaTestCorpus').iterdir()):
        n_docs = len(list(path.glob('*.txt')))
        docs[genre[path.name]] += n_docs
        total_docs += n_docs

        n_tokens = count_hebrew_tokens([str(path)])
        tokens[genre[path.name]] += n_tokens
        total_tokens += n_tokens

    print(r"            & Genre      & Sources        & \# Docs & \# Tokens \\")
    print(r"\midrule")
    for genre in [DICTA, YANSHUF, BOOKS, OFFICIAL, NEWS_MAG, USER_GEN, WIKI]:
        print(f'{genre} & {docs[genre]:6,} & {tokens[genre]:10,} \\\\')
    print(r"\midrule")
    print(f"            & Total      &                & {total_docs:6,} & {total_tokens:10,} \\\\")


if __name__ == '__main__':
    # table_2()
    import train
    # print(count_hebrew_tokens(['hebrew_diacritized/validation']))
    # print(count_hebrew_tokens(train.Full.corpus["premodern"]))
    print(count_hebrew_tokens(train.Full.corpus["automatic"]))
    # print(count_hebrew_tokens(train.Full.corpus["modern"]))
    # for k, v in Counter(remove_niqqud(str(token)) for token in tokens).items():
    #     print(k, v)


from __future__ import annotations
from typing import Iterator, Optional
from pathlib import Path
from dataclasses import dataclass

import numpy as np
import prettytable as pt

from nakdimon import external_apis
from nakdimon import hebrew


@dataclass(frozen=True)
class Document:
    name: str
    system: str
    text: str

    def sentences(self):
        return [sent + '.' for sent in self.text.split('. ') if len(hebrew.remove_niqqud(sent)) > 15]

    def hebrew_items(self) -> list[hebrew.HebrewItem]:
        return list(hebrew.iterate_dotted_text(self.text))

    def tokens(self) -> list[hebrew.Token]:
        return hebrew.tokenize(self.hebrew_items(), strip_nonhebrew=False)

    def vocalized_tokens(self) -> list[hebrew.Token]:
        return [x.vocalize() for x in self.tokens()]


@dataclass(frozen=True)
class DocumentPack:
    source: str
    name: str
    docs: dict[str, Document]

    def __getitem__(self, item):
        return self.docs[item]

    @property
    def expected(self):
        return self.docs['expected']

    @property
    def actual(self):
        assert len(self.docs) == 2
        assert 'expected' in self.docs
        return self.docs[(set(self.docs.keys()) - {'expected'}).pop()]


def system_path_from_expected(path_to_expected: Path, system: str) -> Path:
    return Path(str(path_to_expected).replace('expected', system))


def read_document(system: str, path: Path) -> Document:
    return Document(path.name, system, ' '.join(path.read_text(encoding='utf8').strip().split()))


def read_document_pack(path_to_expected: Path, systems: list[str]) -> DocumentPack:
    return DocumentPack(path_to_expected.parent.name, path_to_expected.name,
                        {system: read_document(system, system_path_from_expected(path_to_expected, system))
                         for system in systems})

class Stats:
    def __init__(self, basepath: Path, vocabulary: Optional[external_apis.MajorityDiacritizer]):
        self.basepath = basepath
        self.vocabulary = vocabulary

    def iter_documents(self, systems: list[str]) -> Iterator[DocumentPack]:
        for folder in self.basepath.iterdir():
            for path_to_expected in folder.iterdir():
                yield read_document_pack(path_to_expected, systems)

    def iter_documents_by_folder(self, systems: list[str]) -> Iterator[list[DocumentPack]]:
        for folder in self.basepath.iterdir():
            yield [read_document_pack(path_to_expected, systems) for path_to_expected in folder.iterdir()]

    def collect_failed_tokens(self, doc_pack: DocumentPack, context: int):
        tokens_of = {system: doc_pack[system].tokens() for system in doc_pack.docs}
        for i in range(len(tokens_of['expected'])):
            res = {system: str(tokens_of[system][i]) for system in doc_pack.docs}
            if len(set(res.values())) > 1:
                pre_nonhebrew, _, post_nonhebrew = tokens_of['expected'][i].split_on_hebrew()
                pre = ' '.join(token.to_text() for token in tokens_of['expected'][i-context:i]) + ' ' + pre_nonhebrew
                post = post_nonhebrew + " " + ' '.join(token.to_text() for token in tokens_of['expected'][i+1:i+context+1])
                res = {system: tokens_of[system][i].split_on_hebrew()[1].to_text() for system in doc_pack.docs}
                yield (pre, res, post)

    def metric_cha(self, doc_pack: DocumentPack) -> float:
        """
        Calculate character-level agreement between actual and expected.
        """
        return self.mean_equal((x, y) for x, y in zip(doc_pack.actual.hebrew_items(), doc_pack.expected.hebrew_items())
                               if hebrew.can_any(x.letter))

    def metric_dec(self, doc_pack: DocumentPack) -> float:
        """
        Calculate nontrivial-decision agreement between actual and expected.
        """
        actual_hebrew = doc_pack.actual.hebrew_items()
        expected_hebrew = doc_pack.expected.hebrew_items()

        return self.mean_equal(
           ((x.niqqud, y.niqqud) for x, y in zip(actual_hebrew, expected_hebrew)
            if hebrew.can_niqqud(x.letter)),

           ((x.dagesh, y.dagesh) for x, y in zip(actual_hebrew, expected_hebrew)
            if hebrew.can_dagesh(x.letter)),

           ((x.sin, y.sin) for x, y in zip(actual_hebrew, expected_hebrew)
            if hebrew.can_sin(x.letter)),
        )

    def is_oov(self, word: str) -> bool:
        if self.vocabulary is None:
            return True
        return self.vocabulary.is_oov(word)

    def metric_wor(self, doc_pack: DocumentPack, oov=False) -> float:
        """
        Calculate token-level agreement between actual and expected,
        for tokens containing at least 2 Hebrew letters.
        """
        return self.mean_equal((x, y) for x, y in zip(doc_pack.actual.tokens(), doc_pack.expected.tokens())
                               if x.is_hebrew() and (not oov or self.is_oov(str(x))))

    def metric_voc(self, doc_pack: DocumentPack, oov=False) -> float:
        """
        Calculate token-level agreement over vocalization, between actual and expected,
        for tokens containing at least 2 Hebrew letters.
        """
        return self.mean_equal((x, y) for x, y in zip(doc_pack.actual.vocalized_tokens(), doc_pack.expected.vocalized_tokens())
                               if x.is_hebrew() and (not oov or self.is_oov(str(x))))

    def mean_equal(self, *pair_iterables) -> float:
        total = 0
        acc = 0
        for pair_iterable in pair_iterables:
            pair_iterable = list(pair_iterable)
            total += len(pair_iterable)
            acc += sum(x == y for x, y in pair_iterable)
        if not total:
            return 0
        return acc / total

    def all_metrics(self, doc_pack: DocumentPack) -> dict[str, float]:
        return {
            'dec': self.metric_dec(doc_pack),
            'cha': self.metric_cha(doc_pack),
            'wor': self.metric_wor(doc_pack),
            'voc': self.metric_voc(doc_pack),
            'wor_oov': self.metric_wor(doc_pack, oov=True),
            'voc_oov': self.metric_voc(doc_pack, oov=True),
        }

    def metricwise_mean(self, iterable) -> dict[str, float]:
        items = list(iterable)
        keys = items[0].keys()
        return {
            key: np.mean([item[key] for item in items])
            for key in keys
        }

    def macro_average(self, system: str) -> dict[str, float]:
        return self.metricwise_mean(
            self.metricwise_mean(self.all_metrics(doc_pack) for doc_pack in folder_packs)
            for folder_packs in self.iter_documents_by_folder(['expected', system])
        )

    def micro_average(self, system: str, vocabulary) -> dict[str, float]:
        return self.metricwise_mean(self.all_metrics(doc_pack) for doc_pack in self.iter_documents(['expected', system]))

    def all_failed(self) -> None:
        for doc_pack in self.iter_documents(['expected', 'Morfix', 'Dicta', 'Nakdimon']):
            for pre, ngrams, post in self.collect_failed_tokens(doc_pack, context=3):
                res = "|".join(ngrams.values())
                print(f'{doc_pack.source}|{doc_pack.name}| {pre}|{res}|{post} |')


def format_latex(system, results) -> None:
    print(r'{system} & {dec:.2%} & {cha:.2%} & {wor:.2%} & {voc:.2%} & {wor_oov:.2%} & {voc_oov:.2%} \\'.format(system=system, **results)
          .replace('%', ''))



def adapt_morfix(expected_filename: str) -> None:
    with open(expected_filename, encoding='utf8') as f:
        expected = list(hebrew.iterate_dotted_text(f.read()))
    actual_filename = expected_filename.replace('expected', 'MorfixOriginal')
    print(actual_filename)
    with open(actual_filename, encoding='utf8') as f:
        actual = list(hebrew.iterate_dotted_text(f.read()))
    fixed_actual = []
    i = 0
    for j in range(len(actual)):
        fixed_actual.append(actual[j])
        if j + 1 >= len(actual):
            break
        e = expected[i+1].letter
        a = actual[j+1].letter
        if e != a:
            assert actual[j + 1].letter == expected[i + 2].letter, hebrew.items_to_text(actual[j-15:j+15]) + ' != ' + hebrew.items_to_text(expected[i-15:i+15])
            if e == 'י' or e == 'א':
                fixed_actual.append(hebrew.HebrewItem(e, e, '', '', ''))
            elif e == 'ל':  # רמאללה -> רמאלה
                t = fixed_actual.pop()
                fixed_actual.append(hebrew.HebrewItem(e, e, '', '', ''))
                fixed_actual.append(t)
            elif e == 'ו':
                if actual[j].niqqud == hebrew.Niqqud.KAMATZ and actual[j].letter == 'ו':
                    add = ''
                elif actual[j].niqqud in [hebrew.Niqqud.REDUCED_KAMATZ, hebrew.Niqqud.KAMATZ, hebrew.Niqqud.HOLAM]:
                    add = hebrew.Niqqud.HOLAM
                    fixed_actual[-1] = actual[j]._replace(niqqud='')
                elif actual[j].niqqud in [hebrew.Niqqud.KUBUTZ]:
                    add = hebrew.Niqqud.SHURUK
                    fixed_actual[-1] = actual[j]._replace(niqqud='')
                else:
                    add = ''
                fixed_actual.append(hebrew.HebrewItem(e, e, '', '', add))
            else:
                assert False, hebrew.items_to_text(actual[j-15:j+15])
            i += 1
        i += 1
    fixed_actual_filename = Path(expected_filename.replace('expected', 'Morfix'))
    fixed_actual_filename.parent.mkdir(parents=True, exist_ok=True)
    with open(fixed_actual_filename, 'w', encoding='utf8') as f:
        print(hebrew.items_to_text(fixed_actual), file=f)


def main(*, test_set: str, systems: list[str]) -> None:
    external_apis.SYSTEMS.update(external_apis.prepare_majority())

    maj = external_apis.MAJ_ALL_WITH_DICTA
    if test_set == 'tests/dicta':
        # Remove systems that depend on the dicta test set for their training
        maj = external_apis.MAJ_ALL_NO_DICTA
        if external_apis.MAJ_ALL_WITH_DICTA in systems:
            del systems[systems.index(external_apis.MAJ_ALL_WITH_DICTA)]
        if 'Nakdimon' in systems:
            del systems[systems.index('Nakdimon')]
    elif test_set == 'tests/new':
        if external_apis.MAJ_ALL_NO_DICTA in systems:
            del systems[systems.index(external_apis.MAJ_ALL_NO_DICTA)]

    stats = Stats(
        basepath=Path(f'{test_set}/expected'),
        vocabulary=external_apis.SYSTEMS[maj]
    )

    table = pt.PrettyTable()
    table.field_names = ['system', 'DEC', 'CHA', 'WOR', 'VOC', 'WOR_OOV', 'VOC_OOV']
    for system in systems:
        results = {k: f'{v:.2%}' for k, v in stats.macro_average(system).items()}
        table.add_row([system, results['dec'], results["cha"], results["wor"], results["voc"], results["wor_oov"], results["voc_oov"]])
    print(table)



# based on https://keras.io/examples/nlp/ner_transformers/

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers


class TransformerBlock(layers.Layer):
    def __init__(self, embed_dim, num_heads, ff_dim, rate=0.1):
        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.ff_dim = ff_dim
        self.rate = rate
        self.att = keras.layers.MultiHeadAttention(
            num_heads=num_heads, key_dim=embed_dim
        )
        self.ffn = keras.Sequential(
            [
                keras.layers.Dense(ff_dim, activation="relu"),
                keras.layers.Dense(embed_dim),
            ]
        )
        self.layernorm1 = keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = keras.layers.LayerNormalization(epsilon=1e-6)
        self.dropout1 = keras.layers.Dropout(rate)
        self.dropout2 = keras.layers.Dropout(rate)

    def get_config(self):
        config = super().get_config()
        config.update({
            "embed_dim": self.embed_dim,
            "num_heads": self.num_heads,
            "ff_dim": self.ff_dim,
            "rate": self.rate,
        })
        return config

    def call(self, inputs, training=False):
        attn_output = self.att(inputs, inputs)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(inputs + attn_output)
        ffn_output = self.ffn(out1)
        ffn_output = self.dropout2(ffn_output, training=training)
        return self.layernorm2(out1 + ffn_output)


class TokenAndPositionEmbedding(layers.Layer):
    def __init__(self, maxlen, vocab_size, embed_dim):
        super().__init__()
        self.maxlen = maxlen
        self.vocab_size = vocab_size
        self.embed_dim = embed_dim
        self.token_emb = keras.layers.Embedding(
            input_dim=vocab_size, output_dim=embed_dim, mask_zero=True
        )
        self.pos_emb = keras.layers.Embedding(input_dim=maxlen, output_dim=embed_dim)

    def get_config(self):
        config = super().get_config()
        config.update({
            "maxlen": self.maxlen,
            "vocab_size": self.vocab_size,
            "embed_dim": self.embed_dim,
        })
        return config

    def call(self, inputs):
        maxlen = tf.shape(inputs)[-1]
        positions = tf.range(start=0, limit=maxlen, delta=1)
        position_embeddings = self.pos_emb(positions)
        token_embeddings = self.token_emb(inputs)
        return token_embeddings + position_embeddings


class Transformer(layers.Layer):
    def __init__(self, num_tags, vocab_size, maxlen, embed_dim, num_heads, ff_dim):
        super().__init__()
        self.embedding_layer = TokenAndPositionEmbedding(maxlen, vocab_size, embed_dim)
        self.transformer_block = TransformerBlock(embed_dim, num_heads, ff_dim)
        self.dropout1 = layers.Dropout(0.1)
        self.ff = layers.Dense(ff_dim, activation="relu")

    def call(self, inputs, training=False):
        x = self.embedding_layer(inputs)
        x = self.transformer_block(x)
        x = self.dropout1(x, training=training)
        x = self.ff(x)
        return x




from __future__ import annotations

import logging
from collections import defaultdict, Counter
import re
import json
from functools import wraps

import requests

from nakdimon.hebrew import Niqqud
from nakdimon import hebrew


class DottingError(RuntimeError):
    pass


def split_string_by_length(text: str, maxlen) -> list[str]:
    return [''.join(s).strip() for s in hebrew.split_by_length(text, maxlen)]


def piecewise(maxlen):
    def inner(fetch):
        @wraps(fetch)
        def fetcher(text):
            return ' '.join(fetch(chunk) for chunk in split_string_by_length(text, maxlen))
        return fetcher
    return inner


def fix_snopi(dotted_text: str, undotted_text: str) -> str:
    items = list(hebrew.iterate_dotted_text(dotted_text))
    for i in range(len(undotted_text)):
        if len(items) == i:
            items.insert(i, hebrew.HebrewItem(undotted_text[i], '', '', '', ''))
        elif undotted_text[i] != ' ' and items[i].letter == ' ':
            del items[i]
        elif undotted_text[i] != items[i].letter:
            items.insert(i, hebrew.HebrewItem(undotted_text[i], '', '', '', ''))
    return hebrew.items_to_text(items)


@piecewise(75)  # estimated maximum for reasonable time
def fetch_snopi(undotted_text: str) -> str:
    # Add bogus continuation in case there's only a single word
    # so Snopi will not decide to answer with single-word-analysis
    logging.debug(repr(undotted_text))
    dummy = False
    if ' ' not in undotted_text:
        dummy = True
        undotted_text = undotted_text + ' 1'

    url = 'http://www.nakdan.com/GetResult.aspx'

    payload = {
        "txt": undotted_text,
        "ktivmale": 'true',
    }
    headers = {
        'Referer': 'http://www.nakdan.com/nakdan.aspx',
    }

    r = requests.post(url, data=payload, headers=headers)
    r.raise_for_status()

    dotted_text = r.text.strip().split('Result')[1][1:-2]
    logging.debug(repr(dotted_text))
    if hebrew.remove_niqqud(dotted_text) != undotted_text:
        logging.debug('Fixing...')
        dotted_text = fix_snopi(dotted_text, undotted_text)
        logging.debug(repr(dotted_text))
    assert hebrew.remove_niqqud(dotted_text) == undotted_text, f'{repr(dotted_text)}\n!=\n{repr(undotted_text)}'
    if dummy:
        dotted_text = dotted_text[:-2]
    return dotted_text


@piecewise(100)
def fetch_morfix(text: str) -> str:
    import sys
    print('Error: Morfix does not allow automatic diacritization! exiting.', file=sys.stderr)
    exit(1)

    url = 'https://nakdan.morfix.co.il/nikud/NikudText'

    payload = {
        "text": text,
        "isLogged": 'false',
    }
    headers = {
    }

    r = requests.post(url, data=payload, headers=headers)
    r.raise_for_status()
    return json.loads(r.json()['nikud'])['OutputText']


@piecewise(10000)
def fetch_dicta(text: str) -> str:
    text = '\n'.join(line for line in text.split('\n') if not line.startswith('https') and not line.startswith('#')).strip()
    def extract_word(k):
        if k['options']:
            res = k['options'][0][0]
            res = res.replace('|', '')
            res = res.replace(Niqqud.KUBUTZ + 'ו' + Niqqud.METEG, 'ו' + Niqqud.SHURUK)
            res = res.replace(Niqqud.HOLAM + 'ו' + Niqqud.METEG, 'ו' + Niqqud.HOLAM)
            res = res.replace(Niqqud.METEG, '')

            res = re.sub(Niqqud.KAMATZ + 'ו' + '(?=[א-ת])', 'ו' + Niqqud.HOLAM, res)
            res = res.replace(Niqqud.REDUCED_KAMATZ + 'ו', 'ו' + Niqqud.HOLAM)

            res = res.replace(hebrew.DAGESH_LETTER * 2, hebrew.DAGESH_LETTER)
            res = res.replace('\u05be', '-')
            res = res.replace('יְהוָֹה', 'יהוה')
            return res
        return k['word']

    url = 'https://nakdan-2-0.loadbalancer.dicta.org.il/api'

    payload = {
        "task": "nakdan",
        "genre": "modern",
        "data": text,
        "addmorph": True,
        "keepqq": False,
        "nodageshdefmem": False,
        "patachma": False,
        "keepmetagim": True,
    }
    headers = {
        'content-type': 'text/plain;charset=UTF-8'
    }

    r = requests.post(url, json=payload, headers=headers)
    r.raise_for_status()
    result = ''.join(extract_word(k) for k in r.json())
    if len(hebrew.find_longest_undotted(result)) > 40:
        raise DottingError('Failed to dot')
    return result


def fetch_dicta_version() -> str:
    url = 'https://nakdan-4-0.loadbalancer.dicta.org.il/api/'
    r = requests.get(url + 'debug')
    r.raise_for_status()
    model_version = r.json()['version'].split("\n")[4].strip()

    r = requests.get(url + 'wlver')
    r.raise_for_status()
    wordlist_version = r.json()['version']

    return f'{model_version}, {wordlist_version}'


def make_fetch_nakdimon(path):
    @piecewise(10000)
    def fetch_nakdimon(text: str) -> str:
        url = 'http://127.0.0.1:5000'

        payload = {
            "text": text,
            "model_name": path,
        }
        headers = {
        }

        r = requests.post(url, data=payload, headers=headers)
        r.raise_for_status()
        return r.text
    return fetch_nakdimon


def make_nakdimon_no_server(path):
    from nakdimon.predict import predict

    def run_nakdimon(text: str) -> str:
        return predict(text, path)

    return run_nakdimon


SYSTEMS = {
    'Snopi': fetch_snopi,  # Too slow
    'Morfix': fetch_morfix,  # terms-of-use issue
    'Dicta': fetch_dicta,
    'Nakdimon': make_nakdimon_no_server('models/Nakdimon.h5'),
}
all_oov = set()


class MajorityDiacritizer:
    dictionary: dict[str, str]

    @staticmethod
    def update_possibilities(possibilities: defaultdict[str, Counter], train_paths: tuple[str, ...]) -> None:
        import hebrew
        for token in hebrew.collect_tokens(train_paths):
            word = token.to_text()
            if word:
                left, word, right = hebrew.split_nonhebrew(word)
                if word:
                    possibilities[hebrew.remove_niqqud(word)][word] += 1
                    for t in token.items:
                        if hebrew.is_hebrew_letter(t.letter):
                            possibilities[t.letter][t.to_text()] += 1

    def __init__(self, possibilities: defaultdict[str, Counter]) -> None:
        self.dictionary = {word: options.most_common(1)[0][0]
                           for word, options in possibilities.items()}

    def is_oov(self, word: str) -> bool:
        left, word, right = hebrew.split_nonhebrew(hebrew.remove_niqqud(word))
        if word not in self.dictionary:
            # all_oov.add(word)
            return True
        return False

    def diacritize_token(self, word: str) -> str:
        left, word, right = hebrew.split_nonhebrew(word)
        word = hebrew.remove_niqqud(word)
        if word in self.dictionary:
            result = self.dictionary[word]
        else:
            result = word  # ''.join([self.dictionary.get(letter, '') for letter in word])
        return left + result + right

    def __call__(self, text: str) -> str:
        return ' '.join([self.diacritize_token(token) for token in hebrew.remove_niqqud(text).split()])


MAJ_MOD = 'MajMod'
MAJ_ALL_NO_DICTA = 'MajAllNoDicta'
MAJ_ALL_WITH_DICTA = 'MajAllWithDicta'


def prepare_majority():
    logging.info('Preparing MajMod...')
    possibilities = defaultdict(Counter)
    res = {}

    MajorityDiacritizer.update_possibilities(possibilities, tuple([
        'hebrew_diacritized/modern/'
    ]))
    res[MAJ_MOD] = MajorityDiacritizer(possibilities)

    logging.info('Preparing MajAllNoDicta...')
    MajorityDiacritizer.update_possibilities(possibilities, tuple([
        'hebrew_diacritized/poetry',
        'hebrew_diacritized/rabanit',
        'hebrew_diacritized/pre_modern',
        'hebrew_diacritized/shortstoryproject_predotted',
        'hebrew_diacritized/shortstoryproject_Dicta',
        'hebrew_diacritized/validation'
    ]))
    res[MAJ_ALL_NO_DICTA] = MajorityDiacritizer(possibilities)

    logging.info('Preparing MajAllWithDicta...')
    MajorityDiacritizer.update_possibilities(possibilities, ('hebrew_diacritized/dictaTestCorpus',))
    res[MAJ_ALL_WITH_DICTA] = MajorityDiacritizer(possibilities)

    logging.info('Done preparing.')
    return res


def fetch_dicta_count_ambiguity(text: str):
    url = 'https://nakdan-2-0.loadbalancer.dicta.org.il/api'

    payload = {
        "task": "nakdan",
        "genre": "modern",
        "data": text,
        # "addmorph": True,
        "keepqq": False,
        "nodageshdefmem": False,
        "patachma": False,
        "keepmetagim": True,
    }
    headers = {
        'content-type': 'text/plain;charset=UTF-8'
    }

    r = requests.post(url, json=payload, headers=headers)
    r.raise_for_status()
    return [len(set(token['options'])) for token in r.json() if not token['sep']]


if __name__ == '__main__':
    SYSTEMS.update(prepare_majority())



import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
import wandb

from nakdimon.dataset import letters_table, NIQQUD_SIZE, DAGESH_SIZE, SIN_SIZE, LETTERS_SIZE
from nakdimon import utils
from nakdimon import hebrew
from nakdimon.train import TrainingParams
from nakdimon import metrics


pretrain_path = f'./models/wiki'
model_name = pretrain_path + 'pretrain.h5'


class BaseModel(keras.Model):
    def __init__(self, units, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.units = units
        self.embed = layers.Embedding(LETTERS_SIZE, units, mask_zero=True)

        self.lstm1 = layers.Bidirectional(layers.LSTM(units, return_sequences=True, dropout=0.1), merge_mode='sum')
        self.lstm2 = layers.Bidirectional(layers.LSTM(units, return_sequences=True, dropout=0.1), merge_mode='sum')

    def call(self, layer):
        layer = self.embed(layer)
        layer = self.lstm1(layer)
        layer = self.lstm2(layer)
        return layer

    def get_config(self):
        return {
            # 'units': self.units,
            'embed': self.embed,
            'lstm1': self.lstm1,
            'lstm2': self.lstm2,
        }

    @classmethod
    def from_config(cls, config, custom_objects=None):
        model = cls(400)  # config['units'])
        model.embed = config['embed']
        model.lstm1 = config['lstm1']
        model.lstm2 = config['lstm2']
        return model


def self_supervized_model(units):
    return keras.Sequential([
        BaseModel(units, name="base"),
        layers.Dense(LETTERS_SIZE)
    ])


class Pretrained(TrainingParams):
    def build_model(self):
        inp = keras.Input(shape=(None,), batch_size=None)

        pretrained = keras.models.load_model(model_name,
                                             custom_objects={'loss': self.loss, 'BaseModel': BaseModel(self.units)})
        base = pretrained.layers[0]

        embed = base.embed(inp)
        layer = base.lstm1(embed)
        layer = base.lstm2(layer)

        layer = pretrained.layers[1](layer)
        outputs = [
            layers.Dense(NIQQUD_SIZE, name='N')(layer),
            layers.Dense(DAGESH_SIZE, name='D')(layer),
            layers.Dense(SIN_SIZE, name='S')(layer),
        ]
        return keras.Model(inputs=inp, outputs=outputs)


class PretrainedModernOnly(Pretrained):
    def epoch_params(self, data):
        lrs = [30e-4, 30e-4, 30e-4, 8e-4, 1e-4]
        yield ('modern', len(lrs), keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs[epoch]))


def get_masked(raw_y, rate):
    mask = np.full(raw_y.shape[0] * raw_y.shape[1], 1)
    mask[:round(len(mask) * rate)] = 0
    np.random.shuffle(mask)
    mask.resize(raw_y.shape, refcheck=False)
    x = raw_y * mask
    y = raw_y  #  * (1 - mask)
    return x, y


def load_plaintext(filename, maxlen):
    import math
    with open(filename, encoding='utf8') as f:
        text = f.read()
    res = np.array([letters_table.char_indices[hebrew.normalize(c)] for c in text])
    res.resize((math.ceil(len(res) / maxlen), maxlen), refcheck=False)
    return res


def pretrain():
    model = self_supervized_model(400)
    model.compile(loss='sparse_categorical_crossentropy',
                  optimizer=keras.optimizers.legacy.Adam(learning_rate=8e-5),
                  metrics='accuracy')

    config = {
        'batch_size': 128,
        'maxlen': 80,
        'units': 400,
        'model': model,
    }

    run = wandb.init(project="dotter",
                     group="pretrain",
                     tags=[],
                     config=config)

    wandb_callback = wandb.keras.WandbCallback(log_batch_frequency=50,
                                               save_model=False,
                                               log_weights=False)

    with run:
        for fname in utils.iterate_files(["../wikipedia/AA"]):
            name = fname.split('/')[-1]
            raw_y = load_plaintext(fname, 80)
            utils.shuffle_in_unison(raw_y)
            x, y = get_masked(raw_y, 0.3)
            model.fit(x, y, batch_size=128, validation_split=0.1, callbacks=[wandb_callback])
            model.save(f'{pretrain_path}/{name}.h5', save_format='tf')

    model.save(model_name, save_format='tf')
    return model


def train_ablation(params):
    from train import train
    model = train(params)
    model.save(f'./models/ablations/{params.name}.h5')


if __name__ == '__main__':
    mode = ''

    if mode == 'pretrain':
        pretrain()
    # elif mode == 'train_ablation':
    #     train_ablation(PretrainedModernOnly())
    else:
        import ablations
        tf.config.set_visible_devices([], 'GPU')
        model_name = 'PretrainedModernOnly'
        model = tf.keras.models.load_model(f'models/ablations/{model_name}.h5',
                                           custom_objects={'loss': TrainingParams().loss})
        print(model_name, *metrics.metricwise_mean(ablations.calculate_metrics(model)).values(), sep=', ')




from __future__ import annotations
from typing import Optional
import logging
from pathlib import Path

import numpy as np
from tensorflow.keras import layers
import tensorflow as tf
import wandb

from nakdimon import dataset
from nakdimon.dataset import NIQQUD_SIZE, DAGESH_SIZE, SIN_SIZE, LETTERS_SIZE
from nakdimon import schedulers
from nakdimon import transformer

# assert tf.config.list_physical_devices('GPU')

VALIDATION_PATH = 'hebrew_diacritized/validation/modern'
MAXLEN = 80


def masked_metric(v, y_true):
    mask = tf.math.not_equal(y_true, 0)
    return tf.reduce_sum(tf.boolean_mask(v, mask)) / tf.cast(tf.math.count_nonzero(mask), tf.float32)


def accuracy(y_true, y_pred):
    return masked_metric(tf.keras.metrics.sparse_categorical_accuracy(y_true, y_pred), y_true)


class NakdimonParams:
    @property
    def name(self):
        return f'{type(self).__name__}({self.units})'

    batch_size = 128
    units = 400
    corpus = {
        'premodern': tuple([
            'hebrew_diacritized/poetry',
            'hebrew_diacritized/rabanit',
            'hebrew_diacritized/pre_modern',
            'hebrew_diacritized/shortstoryproject_predotted'
        ]),
        'automatic': tuple([
            'hebrew_diacritized/shortstoryproject_Dicta',
        ]),
        'modern': tuple([
            'hebrew_diacritized/modern',
            'hebrew_diacritized/dictaTestCorpus',
            'hebrew_diacritized/validation'
        ])
    }

    validation_rate = 0

    subtraining_rate = {'premodern': 1, 'modern': 1}

    def loss(self, y_true, y_pred):
        return masked_metric(tf.keras.losses.sparse_categorical_crossentropy(y_true, y_pred, from_logits=True), y_true)

    def epoch_params(self, data):
        yield ('premodern', 1, schedulers.CircularLearningRate(3e-3, 8e-3, 1e-4, data['premodern'][0], self.batch_size))
        lrs1 = [30e-4, 10e-4]
        yield ('automatic', len(lrs1), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs1[epoch-1]))
        lrs2 = [10e-4, 10e-4, 3e-4]
        yield ('modern', len(lrs2), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs2[epoch-len(lrs1)-1]))

    def deterministic(self):
        np.random.seed(2)
        tf.random.set_seed(2)

    def build_model(self):
        # self.deterministic()

        inp = tf.keras.Input(shape=(None,), batch_size=None)
        embed = layers.Embedding(LETTERS_SIZE, self.units, mask_zero=True)(inp)

        layer = layers.Bidirectional(layers.LSTM(self.units, return_sequences=True, dropout=0.1), merge_mode='sum')(embed)
        layer = layers.Bidirectional(layers.LSTM(self.units, return_sequences=True, dropout=0.1), merge_mode='sum')(layer)
        layer = layers.Dense(self.units)(layer)

        outputs = [
            layers.Dense(NIQQUD_SIZE, name='N')(layer),
            layers.Dense(DAGESH_SIZE, name='D')(layer),
            layers.Dense(SIN_SIZE, name='S')(layer),
        ]
        return tf.keras.Model(inputs=inp, outputs=outputs)

    def initialize_weights(self, model):
        return model


class TrainingParams(NakdimonParams):
    corpus = {
        'premodern': tuple([
            'hebrew_diacritized/poetry',
            'hebrew_diacritized/rabanit',
            'hebrew_diacritized/pre_modern',
            'hebrew_diacritized/shortstoryproject_predotted'
        ]),
        'automatic': tuple([
            'hebrew_diacritized/shortstoryproject_Dicta',
        ]),
        'modern': tuple([
            'hebrew_diacritized/modern',
            'hebrew_diacritized/dictaTestCorpus',
        ])
    }

    def __init__(self, units=NakdimonParams.units):
        self.units = units


class Transformer(TrainingParams):
    batch_size = 64

    def build_model(self) -> tf.keras.Model:
        from tensorflow.keras import layers

        inp = tf.keras.Input(shape=(None,), batch_size=None)

        layer = transformer.TokenAndPositionEmbedding(80, LETTERS_SIZE, self.units)(inp)
        layer = transformer.TransformerBlock(embed_dim=self.units, num_heads=4, ff_dim=self.units, rate=0.1)(layer)
        layer = layers.Dense(self.units, activation='relu')(layer)
        layer = transformer.TransformerBlock(embed_dim=self.units, num_heads=4, ff_dim=self.units, rate=0.1)(layer)
        layer = layers.Dense(self.units, activation='relu')(layer)
        outputs = [
            layers.Dense(NIQQUD_SIZE, name='N')(layer),
            layers.Dense(DAGESH_SIZE, name='D')(layer),
            layers.Dense(SIN_SIZE, name='S')(layer),
        ]
        return tf.keras.Model(inputs=inp, outputs=outputs)

    def epoch_params(self, data):
        lrs0 = [20e-4, 5e-4]
        yield ('premodern', len(lrs0), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs0[epoch]))
        lrs1 = [10e-4, 10e-4]
        yield ('automatic', len(lrs1), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs1[epoch-len(lrs0)]))
        lrs2 = [10e-4] * 6
        yield ('modern', len(lrs2), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs2[epoch-len(lrs0)-len(lrs1)]))


class TwoLevelLSTM(TrainingParams):
    batch_size = 128

    def build_model(self) -> tf.keras.Model:
        from tensorflow.keras import layers

        inp = tf.keras.Input(shape=(None,), batch_size=None)

        layer = layers.Embedding(LETTERS_SIZE, self.units, mask_zero=True)(inp)
        layer = layers.Bidirectional(layers.LSTM(self.units, return_sequences=True, dropout=0.1), merge_mode='sum')(layer)
        layer = layers.Bidirectional(layers.LSTM(self.units, return_sequences=True, dropout=0.1), merge_mode='sum')(layer)
        outputs = [
            layers.Dense(NIQQUD_SIZE, name='N')(layer),
            layers.Dense(DAGESH_SIZE, name='D')(layer),
            layers.Dense(SIN_SIZE, name='S')(layer),
        ]
        return tf.keras.Model(inputs=inp, outputs=outputs)

    def epoch_params(self, data):
        lrs0 = [3e-4]
        yield ('premodern', 1, schedulers.CircularLearningRate(1e-4, 8e-3, 1e-4, data['premodern'][0], self.batch_size))
        lrs1 = [30e-4, 10e-4]
        yield ('automatic', len(lrs1), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs1[epoch-len(lrs0)]))
        lrs2 = [10e-4, 10e-4, 3e-4]
        yield ('modern', len(lrs2), tf.keras.callbacks.LearningRateScheduler(lambda epoch, lr: lrs2[epoch-len(lrs1)-len(lrs0)]))


def get_xy(d: dataset.Data):
    x = d.normalized
    y = {'N': d.niqqud, 'D': d.dagesh, 'S': d.sin}
    return (x, y)


def load_data(params: NakdimonParams):
    logging.info("Loading training data...")
    train_dict = {}
    for stage_name, stage_dataset_filenames in params.corpus.items():
        logging.info(f"Loading training data: {stage_name}...")
        data = dataset.load_data(tuple(stage_dataset_filenames), maxlen=MAXLEN)
        data.shuffle()
        train_dict[stage_name] = get_xy(data)
        logging.info(f"{stage_name} loaded.")
    logging.info("Training data loaded.")
    return train_dict


def load_validation_data():
    logging.info("Loading validation data...")
    data = dataset.load_data(tuple([VALIDATION_PATH]), maxlen=MAXLEN)
    data.shuffle()
    result = get_xy(data)
    logging.info("Validation data loaded.")
    return result


def ablation_metrics(model):
    from nakdimon import predict
    from nakdimon import metrics
    from nakdimon import hebrew

    def calculate_metrics(model, validation_path):
        for path in Path(validation_path).glob('*'):
            logging.debug(path)
            doc = metrics.read_document('expected', path)
            yield metrics.all_metrics(metrics.DocumentPack(
                path.parent.name,
                path.name,
                {
                    'expected': doc,
                    'actual': metrics.Document('actual', 'Nakdimon', predict.predict(hebrew.remove_niqqud(doc.text), model, maxlen=MAXLEN))
                }
            ))

    return metrics.metricwise_mean(calculate_metrics(model, VALIDATION_PATH))


def train(params: NakdimonParams, group, ablation=False, wandb_enabled=False):
    train_dict = load_data(params)
    validation_data = load_validation_data() if ablation else None
    logging.info("Creating model...")
    model = params.build_model()
    model.compile(loss=params.loss,
                  optimizer=tf.keras.optimizers.legacy.Adam(learning_rate=1e-3),
                  metrics=accuracy)

    config = {
        'batch_size': params.batch_size,
        'units': params.units,
        'model': model,
    }

    run = wandb.init(project="dotter",
                     group=group,
                     name=params.name,
                     tags=[],
                     config=config,
                     mode="enabled" if wandb_enabled else "disabled")

    model = params.initialize_weights(model)

    with run:
        last_epoch = 0
        for phase, (stage, n_epochs, scheduler) in enumerate(params.epoch_params(train_dict)):
            logging.info(f"Training phase {phase}: {stage}, {n_epochs} epochs.")

            training_data = (x, y) = train_dict[stage]

            wandb_callback = wandb.keras.WandbCallback(log_batch_frequency=10,
                                                       training_data=training_data,
                                                       validation_data=validation_data,
                                                       save_model=False,
                                                       log_weights=False,
                                                       save_graph=False)

            model.fit(x, y, validation_data=validation_data,
                      initial_epoch=last_epoch,
                      epochs=last_epoch + n_epochs,
                      batch_size=params.batch_size,  # verbose=2,
                      callbacks=[wandb_callback, scheduler])
            last_epoch += n_epochs
            if ablation:
                wandb.log({'epoch': last_epoch, **ablation_metrics(model)})
        if ablation:
            wandb.log({'final': 1, **ablation_metrics(model)})
    return model


def train_ablation(params, group):
    model = train(params, group, ablation=True)
    model.save(f'./models/ablations/{params.name}.h5')


class Full(NakdimonParams):
    validation_rate = 0


def main(*, model_path: str, wandb: bool, ablation_name: Optional[str]):
    if ablation_name is not None:
        import ablations
        params = vars(ablations)[ablation_name]()
        model = train(params, ablation_name, ablation=False, wandb_enabled=wandb)
    else:
        model = train(Full(), 'Full', ablation=False, wandb_enabled=wandb)
    model.save(model_path)



from __future__ import annotations
import collections
import logging
import sys
from pathlib import Path

import requests

from nakdimon import external_apis
from nakdimon import utils
from nakdimon import hebrew


def diacritize_all(system: str, basepath: str, skip_existing: bool, model_path: str) -> None:
    assert system in external_apis.SYSTEMS
    if system == 'Nakdimon':
        path = Path(model_path)
        assert path.suffix == '.h5', f"Expected a path to a h5 file, got {path.suffix}"
        assert path.is_file(), "Expected a path to a h5 file"
        diacritizer = external_apis.make_nakdimon_no_server(model_path)
        system = path.stem
    else:
        diacritizer = external_apis.SYSTEMS[system]

    def diacritize_this(filename: str) -> None:
        infile = Path(filename)
        outfile = Path(filename.replace('expected', system))
        if outfile.exists():
            if skip_existing:
                logging.info(f'{outfile} already exists, skipping')
                return
            else:
                outfile.unlink()
        outfile.parent.mkdir(parents=True, exist_ok=True)
        with open(infile, 'r', encoding='utf8') as f:
            expected = f.read()
        cleaned = hebrew.remove_niqqud(expected)
        actual = diacritizer(cleaned)
        logging.debug(f'Writing {outfile}')
        with open(outfile, 'w', encoding='utf8') as f:
            f.write(actual)
        logging.info(f'Diacritized: {outfile}')

    logging.info(f'Iterating over {basepath}')
    for filename in utils.iterate_files([basepath]):
        logging.info(f'Diacritizing {filename}')
        try:
            diacritize_this(filename)
        except external_apis.DottingError:
            logging.warning("Failed to dot")


def count_all_ambiguity(basepath: str) -> None:
    c = collections.Counter()
    for filename in utils.iterate_files([basepath]):
        print(filename, end=' ' * 30 + '\r', flush=True)

        with open(filename, 'r', encoding='utf8') as r:
            expected = r.read()

        cleaned = hebrew.remove_niqqud(expected)
        actual = external_apis.fetch_dicta_count_ambiguity(cleaned)
        c.update(actual)

    with open('count_ambiguity.txt', 'w', encoding='utf8') as f:
        print(c, file=f)


def main(system: str, test_set: str, skip_existing: bool, model_path: str) -> None:
    if 'Maj' in system:
        external_apis.SYSTEMS.update(external_apis.prepare_majority())
    try:
        diacritize_all(system, f'{test_set}/expected', skip_existing, model_path)
    except requests.exceptions.ConnectionError as ex:
        logging.error(str(ex))
        if system == 'Nakdimon':
            print("Error: Could not connect to Nakdimon. Make sure the Nakdimon server is running.", file=sys.stderr)
        exit(1)



from __future__ import annotations
import flask
import werkzeug
import logging

from nakdimon import predict

app = flask.Flask(__name__)


@app.route('/', methods=["POST"])
def diacritize():
    text = flask.request.values.get('text')
    if not text:
        text = flask.request.files.get('text').read().decode('utf-8')
        if not text:
            raise werkzeug.exceptions.BadRequest
    logging.info(f'request: {text}')
    model_name = flask.request.values.get('model_name')
    logging.info(f'model_name: {model_name}')
    actual = predict.predict(text, predict.load_cached_model(model_name))
    logging.debug(f'result: {actual}')
    response = flask.make_response(actual, 200)
    response.mimetype = "text/plain"
    return response


def main():
    logging.info("Loading models/Nakdimon.h5")
    try:
        predict.predict("שלום", 'models/Nakdimon.h5')
        logging.info("Done loading.")
    except OSError:
        logging.warning("Could not load default model")
    app.run(host='localhost')


if __name__ == '__main__':
    main()



import logging
from functools import lru_cache

import tensorflow as tf

from nakdimon import utils, dataset, hebrew


if tf.config.set_visible_devices([], 'GPU'):
    logging.warning('No GPU available.')


@lru_cache()
def load_cached_model(m):
    model = tf.keras.models.load_model(m, custom_objects={'loss': None})
    return model


def merge_unconditional(texts, tnss, nss, dss, sss):
    res = []
    for ts, tns, ns, ds, ss in zip(texts, tnss, nss, dss, sss):
        sentence = []
        for t, tn, n, d, s in zip(ts, tns, ns, ds, ss):
            if tn == 0:
                break
            sentence.append(t)
            sentence.append(dataset.dagesh_table.indices_char[d] if hebrew.can_dagesh(t) else '\uFEFF')
            sentence.append(dataset.sin_table.indices_char[s] if hebrew.can_sin(t) else '\uFEFF')
            sentence.append(dataset.niqqud_table.indices_char[n] if hebrew.can_niqqud(t) else '\uFEFF')
        res.append(''.join(sentence))
    return res


def predict(text: str, model_or_model_path: tf.Module|str = 'models/Nakdimon.h5', maxlen=10000) -> str:
    if isinstance(model_or_model_path, str):
        model = load_cached_model(model_or_model_path)
    elif isinstance(model_or_model_path, tf.Module):
        model = model_or_model_path
    else:
        raise TypeError(f'Expected str or tf.Module, got {type(model_or_model_path)}')
    data = dataset.Data.from_text(hebrew.iterate_dotted_text(text), maxlen)
    prediction = model.predict(data.normalized)
    [actual_niqqud, actual_dagesh, actual_sin] = [dataset.from_categorical(prediction[0]), dataset.from_categorical(prediction[1]), dataset.from_categorical(prediction[2])]
    actual = merge_unconditional(data.text, data.normalized, actual_niqqud, actual_dagesh, actual_sin)
    return ' '.join(actual).replace('\ufeff', '').replace('  ', ' ').replace(hebrew.RAFE, '')


def main(input_path='-', output_path='-'):
    with utils.smart_open(input_path, 'r', encoding='utf-8') as f:
        text = hebrew.remove_niqqud(f.read())
    text = predict(text)
    with utils.smart_open(output_path, 'w', encoding='utf-8') as f:
        f.write(text)



from __future__ import annotations

import random
import numpy as np

from nakdimon import hebrew, utils



class CharacterTable:
    MASK_TOKEN = ''

    def __init__(self, chars: list[str]) -> None:
        # make sure to be consistent with JS
        self.chars = [CharacterTable.MASK_TOKEN] + chars
        self.char_indices = dict((c, i) for i, c in enumerate(self.chars))
        self.indices_char = dict((i, c) for i, c in enumerate(self.chars))

    def __len__(self) -> int:
        return len(self.chars)

    def to_ids(self, css):
        return [
            [self.char_indices[c] for c in cs] for cs in css
        ]

    def __repr__(self) -> str:
        return repr(self.chars)


letters_table = CharacterTable(hebrew.SPECIAL_TOKENS + hebrew.VALID_LETTERS)
dagesh_table = CharacterTable(hebrew.DAGESH)
sin_table = CharacterTable(hebrew.NIQQUD_SIN)
niqqud_table = CharacterTable(hebrew.NIQQUD)

LETTERS_SIZE = len(letters_table)
NIQQUD_SIZE = len(niqqud_table)
DAGESH_SIZE = len(dagesh_table)
SIN_SIZE = len(sin_table)


def print_tables():
    print('const ALL_TOKENS =', letters_table.chars, end=';\n')
    print('const niqqud_array =', niqqud_table.chars, end=';\n')
    print('const dagesh_array =', dagesh_table.chars, end=';\n')
    print('const sin_array =', sin_table.chars, end=';\n')

    
def from_categorical(t):
    return np.argmax(t, axis=-1)


def merge(texts, tnss, nss, dss, sss):
    res = []
    for ts, tns, ns, ds, ss in zip(texts, tnss, nss, dss, sss):
        sentence = []
        for t, tn, n, d, s in zip(ts, tns, ns, ds, ss):
            if tn == 0:
                break
            sentence.append(t)
            if hebrew.can_dagesh(t):
                sentence.append(dagesh_table.indices_char[d].replace(hebrew.RAFE, ''))
            if hebrew.can_sin(t):
                sentence.append(sin_table.indices_char[s].replace(hebrew.RAFE, ''))
            if hebrew.can_niqqud(t):
                sentence.append(niqqud_table.indices_char[n].replace(hebrew.RAFE, ''))
        res.append(''.join(sentence))
    return res


class Data:
    text: np.ndarray = None
    normalized: np.ndarray = None
    dagesh: np.ndarray = None
    sin: np.ndarray = None
    niqqud: np.ndarray = None

    filenames: tuple[str, ...] = ()

    @staticmethod
    def concatenate(others) -> Data:
        self = Data()
        self.text = np.concatenate([x.text for x in others])
        self.normalized = np.concatenate([x.normalized for x in others])
        self.dagesh = np.concatenate([x.dagesh for x in others])
        self.sin = np.concatenate([x.sin for x in others])
        self.niqqud = np.concatenate([x.niqqud for x in others])
        return self

    def shapes(self) -> tuple[tuple[int, ...], ...]:
        return self.text.shape, self.normalized.shape, self.dagesh.shape, self.sin.shape, self.niqqud.shape

    def shuffle(self) -> None:
        utils.shuffle_in_unison(
            self.text,
            self.normalized,
            self.dagesh,
            self.niqqud,
            self.sin
        )

    @staticmethod
    def from_text(heb_items, maxlen: int) -> Data:
        assert heb_items
        data = Data()
        text, normalized, dagesh, sin, niqqud = zip(*(zip(*line) for line in hebrew.split_by_length(heb_items, maxlen)))

        def pad(ords, dtype='int32', value=0):
            return utils.pad_sequences(ords, maxlen=maxlen, dtype=dtype, value=value)

        data.normalized = pad(letters_table.to_ids(normalized))
        data.dagesh = pad(dagesh_table.to_ids(dagesh))
        data.sin = pad(sin_table.to_ids(sin))
        data.niqqud = pad(niqqud_table.to_ids(niqqud))
        data.text = pad(text, dtype='<U1', value=0)
        return data

    def __len__(self) -> int:
        return self.normalized.shape[0]

    def print_stats(self) -> None:
        print(self.shapes())


def read_corpora(base_paths) -> tuple[tuple[str, list[hebrew.HebrewItem]], ...]:
    return tuple([(filename, list(hebrew.iterate_file(filename))) for filename in utils.iterate_files(base_paths)])


def load_data(base_paths, maxlen: int) -> Data:
    corpora = read_corpora(base_paths)
    corpus = [(filename, Data.from_text(heb_items, maxlen)) for (filename, heb_items) in corpora]
    cs = [c for (_, c) in corpus]
    random.shuffle(cs)
    return Data.concatenate(cs)


if __name__ == '__main__':
    # data = Data.concatenate([Data.from_text(x, maxlen=64) for x in read_corpora(['hebrew_diacritized/modern/wiki/1.txt'])])
    # data.print_stats()
    # print(np.concatenate([data.normalized[:1], data.sin[:1]]))
    # res = merge(data.text[:1], data.normalized[:1], data.niqqud[:1], data.dagesh[:1], data.sin[:1])
    # print(res)
    print_tables()
    print(letters_table.to_ids(["שלום"]))



import argparse
import sys
import os
import logging


def do_train(**kwargs) -> None:
    from nakdimon import train
    train.main(**kwargs)


def do_run_test(**kwargs) -> None:
    from nakdimon import run_test
    run_test.main(**kwargs)


def do_metrics(**kwargs) -> None:
    from nakdimon import metrics
    metrics.main(**kwargs)


def do_predict(**kwargs) -> None:
    from nakdimon import predict
    predict.main(**kwargs)


def do_server(**kwargs):
    import os
    import sys
    import pkgutil
    package = pkgutil.get_loader("server")
    assert package is not None
    logging.info("Executing flask server...")
    os.execv(sys.executable, [sys.executable, package.get_filename()])
    exit(1)


def main() -> None:
    logging.basicConfig(level=logging.INFO, format='%(levelname)s - %(message)s')

    # Parse command line arguments
    parser = argparse.ArgumentParser(
        description="""Train and evaluate Nakdimon and other diacritizers. Reproduce the Nakdimon paper.""",
    )
    parser.add_argument('-q', '--quiet', action='store_true', help='suppress info logging.', default=False)

    subparsers = parser.add_subparsers(help='sub-command help', dest="command", required=True)

    parser_train = subparsers.add_parser('train', help='train Nakdimon')
    parser_train.add_argument('--wandb', action='store_true', help='use wandb.', default=False)
    parser_train.add_argument('--model', help='path to output model (.h5 file)', default='models/Full.h5', dest='model_path')
    parser_train.add_argument('--ablation', help='ablation test', default=None, dest='ablation_name')
    parser_train.set_defaults(func=do_train)

    test_systems = ['Snopi', 'Morfix', 'Dicta', 'Nakdimon', 'MajMod', 'MajAllWithDicta', 'MajAllWithoutDicta']
    # iterate over folders to find available options:
    available_tests = [f'tests/{folder}' for folder in os.listdir('tests/')
                       if os.path.isdir(f'tests/{folder}') and os.path.isdir(f'tests/{folder}/expected')]

    parser_test = subparsers.add_parser('run_test', help='diacritize a test set')
    parser_test.add_argument('--test_set', choices=available_tests, help='choose test set', default='tests/new')
    parser_test.add_argument('--system', choices=test_systems, help='diacritization system to use', default='Nakdimon')
    parser_test.add_argument('--model', help='path to model (.h5 file)', default='models/Nakdimon.h5', dest='model_path')
    parser_test.add_argument('--skip-existing', action='store_true', help='skip existing files')
    parser_test.set_defaults(func=do_run_test)

    parser_predict = subparsers.add_parser('predict', help='diacritize a text file')
    parser_predict.add_argument('input_path', help='input file')
    parser_predict.add_argument('output_path', help='output file')
    parser_predict.set_defaults(func=do_predict)

    # parser_daemon = subparsers.add_parser('server', help='run Nakdimon server as a daemon')
    # parser_daemon.set_defaults(func=do_server)

    parser_eval = subparsers.add_parser('results', help='evaluate the results of a test run')
    parser_eval.add_argument('--test_set', choices=available_tests, help='choose test set', default='tests/new')
    partial_result, _ = parser.parse_known_args()
    if partial_result.command == 'results':
        test_systems = [folder for folder in os.listdir(partial_result.test_set)
                        if os.path.isdir(f'{partial_result.test_set}/{folder}') and folder != 'expected']
    parser_eval.add_argument('--systems', choices=test_systems, nargs='+', help='list of systems to evaluate',
                             default=test_systems)
    parser_eval.set_defaults(func=do_metrics)

    args = parser.parse_args()

    if args.quiet:
        logging.disable(logging.INFO)
    del args.quiet

    kwargs = vars(args).copy()
    del kwargs['command']
    del kwargs['func']
    args.func(**kwargs)

    sys.exit(0)

if __name__ == '__main__':
    main()



from __future__ import annotations
from typing import Iterable

import sys
import contextlib
import os

import numpy as np


def iterate_files(base_paths: Iterable[str]) -> list[str]:
    for name in base_paths:
        if not os.path.isdir(name):
            yield name
            continue
        for root, dirs, files in os.walk(name):
            for fname in files:
                path = os.path.join(root, fname)
                yield path


def read_file(filename):
    with open(filename, 'r', encoding='utf-8') as f:
        return f.read()


# from: https://stackoverflow.com/a/45735618/2289509
@contextlib.contextmanager
def smart_open(filename: str, mode: str = 'r', *args, **kwargs):
    """Open files and i/o streams transparently."""
    if filename == '-':
        if 'r' in mode:
            stream = sys.stdin
        else:
            stream = sys.stdout
        if 'b' in mode:
            fh = stream.buffer
        else:
            fh = stream
        close = False
    else:
        fh = open(filename, mode, *args, **kwargs)
        close = True

    try:
        yield fh
    finally:
        if close:
            try:
                fh.close()
            except AttributeError:
                pass


def pad_sequences(sequences, maxlen, dtype, value) -> np.ndarray:
    # based on keras' pad_sequences()
    num_samples = len(sequences)

    # take the sample shape from the first non empty sequence
    # checking for consistency in the main loop below.
    sample_shape = tuple()
    for s in sequences:
        if len(s) > 0:
            sample_shape = np.asarray(s).shape[1:]
            break

    x = np.full((num_samples, maxlen) + sample_shape, value, dtype=dtype)
    for idx, s in enumerate(sequences):
        if not len(s):
            continue  # empty list/array was found
        trunc = s[:maxlen]
        x[idx, :len(trunc)] = np.asarray(trunc, dtype=dtype)
    return x


def shuffle_in_unison(*arrs):
    rng_state = np.random.get_state()
    for arr in arrs:
        np.random.set_state(rng_state)
        np.random.shuffle(arr)



import os
from collections import Counter

from nakdimon import hebrew


def make_table_2():
    folder_to_source = {
        'blogs': r'Blogs \& Forums',
        'books': r"Books \& Forums",
        'dont_panic': r"Books \& Forums",
        'forums': r'Blogs \& Forums',
        'govil': "gov.il",
        'modernTestCorpus': 'DictaTest',
        'news': 'Online Outlets',
        'random_internet': r'Books \& Forums',
        'wiki': 'he.wikipedia',
        'yanshuf': 'Yanshuf',
        'ynet': 'Online Outlets'
    }
    source_to_type = {
        r"Blogs \& Forums": 'User-gen.',
        r"Books \& Forums": 'Literary',
        "Yanshuf": "News",
        'he.wikipedia': "Wiki",
        'gov.il': 'Official',
        'Online Outlets': 'News / Mag',
        'DictaTest': 'Wiki',
    }

    count_words = Counter()
    count_docs = Counter()
    for folder, source in folder_to_source.items():
        path = f'hebrew_diacritized/modern/{folder}'
        tokens = [
            t for t in hebrew.collect_tokens([path])
            if len(t.strip_nonhebrew().items) > 1
        ]
        count_words[source] += len(tokens)
        count_docs[source] += len(os.listdir(path))
    print(r"& Genre & Sources & \# Docs & \# Tokens \\")
    print(r"\midrule")
    for source in count_words:
        print(rf"& {source_to_type[source]} & {source} & {count_docs[source]} & {count_words[source]} \\")
    print(r" \midrule")
    print(rf"& Total &  & {sum(count_docs.values())} & {sum(count_words.values())} \\")
    print(r" \midrule")


if __name__ == '__main__':
    make_table_2()
